//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//file main.cpp
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "permutation.h"
#include "orientation.h"
#include "oriperm.h"
#include "mappedcube.h"
#include "ida.h"

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>


//+++++++++++++global variables++++++++++++++++++++++++++++++++++++++++++++++
unsigned char genAxis[25]= {255};
unsigned char genPower[25];

char moveConj[N_SYM][N_MOVE];
char symMult[N_SYM][N_SYM];
char symInv[N_SYM];

EOPerm eCubeSym[N_SYM],eCubeSymInv[N_SYM];

unsigned short (*twistConj)[N_SYM];
unsigned short(*twistMove)[N_MOVE];
unsigned short (*sSliceMove)[N_MOVE];
unsigned short (*cPosMove)[N_MOVE];
SymRep (*flipSliceMove)[N_MOVE];

int* flipSliceRepr;

char* phase1Prun;
//+++++++++++++end global variables+++++++++++++++++++++++++++++++++++++++++++


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void InitTables()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	COPerm::InitSymmetryGroupTables();
	COPerm::InitMoveConjugateTable();

	EOPerm::InitCubeSymTable();
	EOPerm::MakeFlipSliceRepTable();

	MappedCube::InitTwistMoveTable();
	MappedCube::InitFlipSliceMoveTable();
	MappedCube::InitCPosMoveTable();
	MappedCube::InitSortedSliceMoveTable();

	MappedCube::InitTwistConjugateTable();

	MappedCube::InitFlipSliceTwistPrunTable();


}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void DeleteTables()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	delete [] twistConj;
	delete [] twistMove;
	delete [] flipSliceMove;
	delete [] cPosMove;
	delete [] sSliceMove;

	delete [] flipSliceRepr;
	delete [] phase1Prun;

}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void main(int argc, char* argv[])
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	
	ifstream f(PHASE1_PRUN,ios::binary|ios::nocreate);
	if (!f)
	{	cout<<"The first time you run the optimizer program, a few tables will be created."<<endl;
		cout<<"This needs about 74MB of disk space and takes less than 30 minutes on a P350."<<endl;
		cout<<"You should have at least 90 MB of RAM,"<<endl;
		cout<<"Proceed (y/n)? ";
		char answer;
		cin>>answer;
		if (answer!='y'&&answer!='Y') return;
	}
	else
	{
		f.close();
	}
	InitTables();
	cout<<endl;
	int n;
	if (argc<3)
	{	cout<<"Usage: optimize [-n] inputFileName outputFileName"<<endl;
		cout<<endl<<"Options:"<<endl;
		cout<<"  -n     compute all sequences which have a length < MINLENGTH + n";
		return;
	}
	bool all=false;
	if (argv[1][0]=='-')
	{	all=true;
		n=argv[1][1]-'0';
		argv[1]=argv[2];
		argv[2]=argv[3];
	}
	ifstream in(argv[1],ios::nocreate);
	ofstream out(argv[2]);
	if (!in)
	{	cout<<"Error: could not open inputfile"<<endl;
		return;
	}
	if (!out)
	{	cout<<"Error: could not open outputfile"<<endl;
		return;
	}
		
	while (true)
	{
		char buffer[256];
		for (int i=0;i<256;i++)
			buffer[i]=(char)255;
		in.getline(buffer,sizeof(buffer));
 		if (!in) break;
		if (buffer[0]==0||buffer[0]<'B') continue;
		Move T;
		COPerm c;
		EOPerm e;
		MappedCube mc(c,e);
		char *p=buffer;
		while (true)
		{	switch (*(p++))
			{	case 'U':T=U1;break;
				case 'R':T=R1;break;
				case 'F':T=F1;break;
				case 'D':T=D1;break;
				case 'L':T=L1;break;
				case 'B':T=B1;break;
				default:T=LAST;
			}
			if (T==LAST) break;
			{	switch (*(p++))
				{	case '\'':
					case '-':
					case '3':mc.Turn(T);mc.Turn(T);mc.Turn(T);break;
					case '2':mc.Turn(T);mc.Turn(T);break;
					case ' ':
					case '\0':
					case '1':mc.Turn(T);break;
				}
			}
			while (*p==' ' || *p=='.') p++;
		}
		
		IDA ida(mc);
		int depth =ida.NextSolution(24);
		ida.PrintInvManeuver(out);
		ida.PrintInvManeuver(cout);
		if (all)
		{
			while (ida.NextSolution(depth+n-1)>0)
			{
				ida.PrintInvManeuver(out);
				ida.PrintInvManeuver(cout);
			}
			ida.PrintSeparation(out);
			ida.PrintSeparation(cout);
		}
	}
}