//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//file mappedcube.cpp
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "mappedcube.h"
# include <fstream.h>

#define EM_ROT EOPerm::M_URF3
#define EM_ROTINV (EOPerm::M_URF3*EOPerm::M_URF3)
#define CM_ROT COPerm::M_URF3
#define CM_ROTINV (COPerm::M_URF3*COPerm::M_URF3)

//the global variables are defined in main.cpp
extern unsigned char genAxis[];
extern unsigned char genPower[];
extern EOPerm eCubeSym[],eCubeSymInv[];
extern char symInv[N_SYM],symMult[N_SYM][N_SYM];
extern char moveConj[N_SYM][N_MOVE];
extern unsigned short (*twistConj)[N_SYM];
extern unsigned short(*twistMove)[N_MOVE];
extern unsigned short (*sSliceMove)[N_MOVE];
extern unsigned short (*cPosMove)[N_MOVE];
extern SymRep (*flipSliceMove)[N_MOVE];
extern int* flipSliceRepr;
extern char* phase1Prun;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MappedCube::MappedCube(COPerm c,EOPerm e)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  
	flipSliceUD=FlipSliceSymRep(e);
	twistUD=c.Twist();
	flipSliceRL=FlipSliceSymRep(EM_ROT*e*EM_ROTINV);
	twistRL=(CM_ROT*c*CM_ROTINV).Twist();
	flipSliceFB=FlipSliceSymRep(EM_ROTINV*e*EM_ROT);
	twistFB=(CM_ROTINV*c*CM_ROT).Twist();

	cPos=c.Pos();
	sSliceUD=e.SortedSlice();
	sSliceRL=(EM_ROT*e*EM_ROTINV).SortedSlice();
	sSliceFB=(EM_ROTINV*e*EM_ROT).SortedSlice();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MappedCube::MappedCube(int flipSliceTwist)
//used to create the pruning table
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	//flipSliceTwist=twist+N_CTWIST*flipSlice.classIdx
	flipSliceUD.sym=0;
	flipSliceUD.classIdx=flipSliceTwist/N_CTWIST;
	twistUD=flipSliceTwist%N_CTWIST;
	flipSliceRL.sym=flipSliceFB.sym=0;
	flipSliceRL.classIdx=flipSliceFB.classIdx=0;

	cPos=0;
	sSliceUD=0;
	sSliceRL=0;
	sSliceFB=0;

}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int MappedCube::FlipSliceTwist()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	//we first conjugate the cube by flipSlice.sym. 
	//this conjugation does the following:
	//flipSlice.classIdx does not change
	//flipSlice.sym -> 0
	//twist -> newTwist
	//
	unsigned short newTwist=twistConj[twistUD][flipSliceUD.sym];
	return newTwist+N_CTWIST*flipSliceUD.classIdx;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MappedCube::MappedCube(unsigned char* genAxis, unsigned char* genPower)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  	
	
	cPos=0;
	sSliceUD=sSliceRL=sSliceFB=0;

	twistUD=twistRL=twistFB=0;
	flipSliceUD.classIdx=flipSliceRL.classIdx=flipSliceFB.classIdx=0;
	flipSliceUD.sym=flipSliceRL.sym=flipSliceFB.sym=0;
	
	int k=0;
	while (genAxis[k]<=B1 && k<N_MAXGEN)
	{	for (char j=0;j<genPower[k];j++)
			Turn((Move)genAxis[k]);
		k++;
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::Turn(Move T)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  char move;
	
	SymRep sr;
	/*
	with s=sr.sym and c=sr.classIdx we have
	(s^-1*c*s)*T=s^-1*c*(s*T*s^-1)*s=s^-1*(c*T)*s=s^-1*(s^-1*c*s)*s=
	(s*s)^-1*c*(s*s)
	*/
	move=moveConj[flipSliceUD.sym][T];//s*T*s^-1=T
	sr=flipSliceMove[flipSliceUD.classIdx][move];//c*T=s^-1*c*s
	flipSliceUD.classIdx=sr.classIdx;
	flipSliceUD.sym=symMult[sr.sym][flipSliceUD.sym];//s*s=s

	twistUD=twistMove[twistUD][T];

	char rlT=moveConj[1][T];//nT=M_URF3*T*M_URF3^-1
	move=moveConj[flipSliceRL.sym][rlT];
	sr=flipSliceMove[flipSliceRL.classIdx][move];
	flipSliceRL.classIdx=sr.classIdx;
	flipSliceRL.sym=symMult[sr.sym][flipSliceRL.sym];

	twistRL=twistMove[twistRL][rlT];

	char fbT=moveConj[2][T];//nT=M_URF3^-1*T*M_URF3
	move=moveConj[flipSliceFB.sym][fbT];
	sr=flipSliceMove[flipSliceFB.classIdx][move];
	flipSliceFB.classIdx=sr.classIdx;
	flipSliceFB.sym=symMult[sr.sym][flipSliceFB.sym];

	twistFB=twistMove[twistFB][fbT];

	cPos=cPosMove[cPos][T];
	sSliceUD=sSliceMove[sSliceUD][T];
	sSliceRL=sSliceMove[sSliceRL][rlT];
	sSliceFB=sSliceMove[sSliceFB][fbT];
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::Print()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	cout<<"flipSliceUD: "<<flipSliceUD.classIdx<<"/"<<(int)flipSliceUD.sym;
	cout<<"  twistUD: "<<twistUD<<" sSliceUD: "<<sSliceUD<<endl;
	cout<<"flipSliceRL: "<<flipSliceRL.classIdx<<"/"<<(int)flipSliceRL.sym;
	cout<<"  twistRL: "<<twistRL<<" sSliceRL: "<<sSliceRL<<endl;
	cout<<"flipSliceFB: "<<flipSliceFB.classIdx<<"/"<<(int)flipSliceFB.sym;
	cout<<"  twistFB: "<<twistFB<<" sSliceFB: "<<sSliceFB<<endl;
	cout<<"cPos: "<<cPos<<endl;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::InitTwistConjugateTable()
//We need this table to compute the FlipSliceTwist() which itself is needed for the
//FlipSliceTwistPrunTable(). With s=flipSlice.sym and c=flipSlice.classIdx, the
//s^-1*c*s describes the state. By conjugation with s we get c, but the twist transforms
//to "s*twist*s^-1" := twistConj[twist][s]
//The conjugation is only allowed for Cube Symmetries, which preserve the UD-axis!
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	int sz=sizeof(unsigned short);
	
	twistConj=(unsigned short(*)[N_SYM])new unsigned short[N_CTWIST*N_SYM];

	ifstream f(TWIST_CONJ,ios::binary|ios::nocreate);
	if (!f)
	{	cout<<"Generating "<<TWIST_CONJ<<endl;
		
		COPerm* cCubeSym = new COPerm[N_SYM];
		COPerm* cCubeSymInv = new COPerm[N_SYM];
		for (int k=0;k<N_SYM;k++)
		{	cCubeSym[k]=COPerm::CubeSym(k);
			if (k<N_SYM/2)
				cCubeSymInv[k]=COPerm::CubeSym(symInv[k]);
			else  //the reflection part must be cancelled here and applied later
				cCubeSymInv[k]=COPerm::CubeSym(symInv[k]-N_SYM/2);
		}

	
		for (int i=0;i<N_CTWIST;i++)
		{	COPerm c(0,i);
			for (char s=0;s<N_SYM;s +=3)//must preserve UD-axis
			{	COPerm ce = cCubeSym[s]*c*cCubeSymInv[s];
				if (s>=N_SYM/2) //apply the cancelled reflection here
					ce=ce*COPerm::M_LR2;
								
				twistConj[i][s]=ce.Twist();
			}
		}
		ofstream of(TWIST_CONJ,ios::binary);
		of.write((char *)twistConj,N_CTWIST*N_SYM*sizeof(unsigned short));
		
		delete[] cCubeSym;
		delete[] cCubeSymInv; 

	}
	else
	f.read((char *)twistConj,N_CTWIST*N_SYM*sizeof(unsigned short));
		
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::InitFlipSliceTwistPrunTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	ifstream f(PHASE1_PRUN,ios::binary|ios::nocreate);
	char* p=new char[N_PHASE1/2];
	
	if (!f)
	{	
		ifstream fsym(FLIPSLICE_SYM,ios::binary|ios::nocreate);
		if (!fsym)
		{	cout<<"Fatal error, cannot open "<<FLIPSLICE_SYM<<endl;
			char t; cin>>t;
		}
		short int* sym= new short int[N_FLIPSLICE];
		fsym.read((char*)sym,N_FLIPSLICE*sizeof(short int));
		
		cout<<"Generating "<<PHASE1_PRUN<<". Maximal depth is 12."<<endl;
		for (int i=0;i<N_PHASE1/2;i++)
			p[i]=(char)(0xff);//even!
	
		COPerm c0;EOPerm e0;
		int base,offset,done=0;
		p[0]=(char)0x0f;//Identity has index 0;
		done++;
		int oldDone=done;

		MappedCube e(0);
		for (char depth=0;depth<14;depth++)
		{	for (int j=0;j<N_PHASE1;j++)
			{	
				base=j/2;offset=j%2;
				char tag=(offset==1?p[base]&(char)0x0f:
							(char)((p[base]&0xf0)>>4));
				if (tag==depth)
				{	for (char T=U1;T<=B1;T++)
					{	//*******MappedCube e(j);**********
						e.flipSliceUD.sym=0;
						e.flipSliceUD.classIdx=j/N_CTWIST;
						e.twistUD=j%N_CTWIST;
						//*********************************
						for (int t=0;t<3;t++)
						{	
							//*************e.Turn((Move)T);************************
							char move=moveConj[e.flipSliceUD.sym][T];
							SymRep sr=flipSliceMove[e.flipSliceUD.classIdx][move];
							e.flipSliceUD.classIdx=sr.classIdx;
							e.flipSliceUD.sym=symMult[sr.sym][e.flipSliceUD.sym];
							e.twistUD=twistMove[e.twistUD][T];
							//*****************************************************
							char symSave=(char)(e.flipSliceUD.sym);
							for (char k=0;k<N_SYM;k +=3)
							{	
								short int bitMask=1;
								bitMask <<=(k/3);
								if ((sym[e.flipSliceUD.classIdx]&bitMask)!=0)
								{//if representant r has symmetry k, we have k^-1*r*k =r and
								 //s^-1*r*s = s^-1*k^-1*r*k*s
									e.flipSliceUD.sym=symMult[k][symSave];
									int idx=e.FlipSliceTwist();
									base=idx/2; offset=idx%2;
									tag=(offset==1?p[base]&(char)0x0f:
											(char)((p[base]&0xf0)>>4));
									if (tag==0x0f)//still empty
									{	p[base]=
										(offset==1?(char)((p[base]&0xf0)|(depth+1)):
											(char)((p[base]&0x0f)|((depth+1)<<4)));
										done +=1;
									}
								}
								if (sym[e.flipSliceUD.classIdx]==1) break;
							}//for k
							e.flipSliceUD.sym=symSave;
						}//for t
					}//for T
				}//if tag
			}//for J
			cout<<"Depth: "<<(int)depth+1<<" "<<"Count: "<<done-oldDone<<endl;
			if (done==oldDone) break;
			oldDone=done;
		}//for depth
	  	delete [] sym;
		ofstream of(PHASE1_PRUN,ios::binary);
		of.write(p,N_PHASE1/2);
		cout<<endl<<"The table generation has terminated."<<endl;
		cout<<"The optimizer is ready for use now."<<endl;
	}//if !f
	else
		f.read(p,N_PHASE1/2);

	phase1Prun=p;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SymRep MappedCube::FlipSliceSymRep(EOPerm e)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	SymRep sr;
		
	int done=0;
	int maxIdx=flipSliceRepr[N_FLIPSLICE-1];

	for (char s=0;s<N_SYM;s +=3)//we preserve the UD-axis
	{	EOPerm e1 = eCubeSym[s]*e*eCubeSymInv[s];
		int idx=e1.FlipSlice();
		if (idx >maxIdx) continue;

		int rightBound=N_FLIPSLICE;//not N_FLIPSLICE-1 !!!
		int leftBound=0;
		int a=-1;
		int aOld;

		do
		{	aOld=a;
			a=(rightBound+leftBound)/2;
			if (idx<flipSliceRepr[a]) rightBound=a;
			else
			{	if (idx>flipSliceRepr[a]) leftBound=a;
				else //so we have transformed to a representant
				{	
					sr.sym=s;	//because s^-1*r*s represents the element
					sr.classIdx=a;
					done=1;
					break;
				}
			}
		}
		while (a!=aOld);
		if (done==1) break;
	}
	if (done==0)
	{	cout<<"Fatal error in MappedCube::FlipSliceSymRep";
		char t;cin>>t;
		return sr;
	}
	else
		return sr;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::InitFlipSliceMoveTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	ifstream f0(FLIPSLICE,ios::binary);
	if (!f0)
	{	cout<<"Fatal error, cannot open "<<FLIPSLICE<<endl;
		char t; cin>>t;
	}
	flipSliceRepr=new int[N_FLIPSLICE];
	f0.read((char *)flipSliceRepr,N_FLIPSLICE*sizeof(int));
	flipSliceMove=(SymRep(*)[N_MOVE])new SymRep[N_FLIPSLICE*N_MOVE];


	ifstream f(FLIPSLICE_MOVE,ios::binary|ios::nocreate);
	if (!f)
	{	
		cout<<"Generating "<<FLIPSLICE_MOVE<<endl;
		for (int i=0;i<N_FLIPSLICE;i++)
		{	
			if (i%10000==0)
				cout<<i<<" of "<<N_FLIPSLICE<<endl;
			int repIdx=flipSliceRepr[i];
			for (char T=U1;T<LAST;T++)
			{	
				EOPerm e(repIdx);
				e.Turn((Move)T);
				flipSliceMove[i][T]=FlipSliceSymRep(e);
			}
		}
		ofstream of(FLIPSLICE_MOVE,ios::binary);
		of.write((char *)flipSliceMove,N_FLIPSLICE*N_MOVE*sizeof(SymRep));
	}
	else
		f.read((char *)flipSliceMove,N_FLIPSLICE*N_MOVE*sizeof(SymRep));

}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::InitTwistMoveTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	
	twistMove=(unsigned short(*)[N_MOVE])new unsigned short[N_CTWIST*N_MOVE];
	ifstream f(CTWIST_MOVE,ios::binary|ios::nocreate);
	if (!f)
	{	cout<<"Generating "<<CTWIST_MOVE<<endl;
		for (int i=0;i<N_CTWIST;i++)
		{
			for (char T=U1;T<LAST;T++)
			{	COPerm c(0,i);
				c.Turn((Move)T);
				twistMove[i][T]=c.Twist();
			}
		}
		ofstream of(CTWIST_MOVE,ios::binary);
		of.write((char *)twistMove,N_CTWIST*N_MOVE*sizeof(unsigned short));
	}
	else
		f.read((char *)twistMove,N_CTWIST*N_MOVE*sizeof(unsigned short));
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::InitSortedSliceMoveTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	sSliceMove=(unsigned short(*)[N_MOVE])new unsigned short[N_SORTSLICE*N_MOVE];
	ifstream f(SORTSLICE_MOVE,ios::binary|ios::nocreate);
	if (!f)
	{	cout<<"Generating "<<SORTSLICE_MOVE<<endl;
		for (short int i=0;i<N_SORTSLICE;i++)
		{	
			for (char T=U1;T<LAST;T++)
			{	EOPerm e(i);
				e.Turn((Move)T);
				sSliceMove[i][T]=e.SortedSlice();
			}
		}
		ofstream of(SORTSLICE_MOVE,ios::binary);
		of.write((char *)sSliceMove,N_SORTSLICE*N_MOVE*sizeof(unsigned short));
	}
	else
		f.read((char *)sSliceMove,N_SORTSLICE*N_MOVE*sizeof(unsigned short));
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void MappedCube::InitCPosMoveTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	cPosMove=(unsigned short(*)[N_MOVE])new unsigned short[N_CPOS*N_MOVE];
	ifstream f(CPOS_MOVE,ios::binary|ios::nocreate);
	if (!f)
	{	cout<<"Generating "<<CPOS_MOVE<<endl;
		for (int i=0;i<N_CPOS;i++)
		{	if (i%10000==0)
				cout<<i<<" of "<<N_CPOS<<endl;
			for (char T=U1;T<LAST;T++)
			{	COPerm c(i,0);
				c.Turn((Move)T);
				cPosMove[i][T]=c.Pos();
			}
		}
		ofstream of(CPOS_MOVE,ios::binary);
		of.write((char *)cPosMove,N_CPOS*N_MOVE*sizeof(unsigned short));
	}
	else
		f.read((char *)cPosMove,N_CPOS*N_MOVE*sizeof(unsigned short));
}
