//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//file oriperm.cpp
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "oriperm.h"
#include <iostream.h>
#include <fstream.h>

extern EOPerm eCubeSym[],eCubeSymInv[];
extern char symInv[N_SYM],symMult[N_SYM][N_SYM];
extern char moveConj[N_SYM][N_MOVE];

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm::OPerm(char size, char order):perm(size),ori(size,order)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm::OPerm(const Perm& p,char order):perm(p),ori(p.size,order)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm::OPerm(const Ori& o):perm(o.size),ori(o)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm::OPerm(const Perm& p, const Ori& o):perm(p),ori(o)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm::OPerm(const OPerm& op):perm(op.perm),ori(op.ori)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm& OPerm::operator=(OPerm& op)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	perm=op.perm;
	ori=op.ori;
	return *this;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm OPerm::operator *(OPerm& p)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	OPerm result(ori.size,ori.order);
	for (char i=0;i<perm.size;i++)
	{	result.perm.data[i]=p.perm.data[perm.data[i]];
		result.ori.data[i]=(char)((ori.data[i]+p.ori.data[perm.data[i]])%ori.order);
	}
	return result;
}//size and order must be the same for both arguments of "*"

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm::OPerm(char size,char order,int pIndex,int oIndex):
	perm(size,pIndex),ori(size,order)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	Ori o(size,order,oIndex);
	OPerm op(perm.Inverse(),o);
	ori=(op.Inverse()).ori;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
OPerm OPerm::Inverse()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  char size=ori.size, order=ori.order;
	OPerm op(size,order);
	for (char i=0;i<size;i++)
	{  char j=perm.data[i];
		op.perm.data[j]=i;
		op.ori.data[j]=(char)((order-ori.data[i])%order);
	}
	return op;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void OPerm::Print()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	perm.Print();
	ori.Print();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int OPerm::PIndex()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	return perm.Index();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int OPerm::OIndex()
//We may not return just ori.Index(), because we need the following property:
//if a,b are OPermutations with a.OIndex() = b.OIndex(), then for any
//OPermutation c we have (a*c).OIndex()=(b*c).OIndex()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  
	OPerm op=Inverse();
	return op.ori.Index();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm():OPerm(N_EDGE,N_FLIP)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(const EPerm& p):OPerm(p,N_FLIP)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(int pIndex, int oIndex):OPerm(N_EDGE,N_FLIP,pIndex,oIndex)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(int flipSlice):OPerm(N_EDGE,N_FLIP)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	
	int slice = flipSlice/N_EFLIP;
	int flip = flipSlice%N_EFLIP;

	int c,n=0;
	for (int m=0;m<4;m++)
	{	while (slice >= (c=Cnk((char)(N_EDGE- ++n),(char)(3-m))) &&c!=0)
			slice -=c;
		char buf=perm.data[m];
		perm.data[m]=(char)(n-1);
		for (int i=0;i<N_EDGE;i++)
		{	if (perm.data[i]==(char)(n-1) && i!=m)
			{	perm.data[i]=buf;
				break;
			}
		}
	}
	Ori o(N_EDGE,N_FLIP,flip);
	OPerm op(perm.Inverse(),o);
	ori=(op.Inverse()).ori;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(short int sortedSlice):OPerm(N_EDGE,N_FLIP)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	int sort = sortedSlice%24;
	int place = sortedSlice/24;
	int order[4];
	char buf;

	int c,n=0;
	for (int m=0;m<4;m++)
	{	while (place >= (c=Cnk((char)(N_EDGE- ++n),(char)(3-m))) &&c!=0)
			place -=c;
		char buf=perm.data[m];
		perm.data[m]=(char)(n-1);
		for (int i=0;i<N_EDGE;i++)
		{	if (perm.data[i]==(char)(n-1) && i!=m)
			{	perm.data[i]=buf;
				break;
			}
		}
	}
	for (int i=2;i<=4;i++)
	{	order[4-i]=sort%i;
		sort /=i;
	}
	for (int j=0;j<3;j++)
	for (int k=order[j];k>0;k--)
	{	buf=perm.data[j+k];
		perm.data[j+k]=perm.data[j+k-1];
		perm.data[j+k-1]=buf;
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(const EOri& eo):OPerm(eo)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(const EPerm& p12, const EOri& eo):OPerm(p12,eo)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(const EOPerm& eop):OPerm(eop)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(const OPerm& op):OPerm(op)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm::EOPerm(unsigned char* genAxis, unsigned char* genPower):
	OPerm(N_EDGE,N_FLIP)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  int i=0;
	while (genAxis[i]<=B1 && i<N_MAXGEN)
	{	for (char j=0;j<genPower[i];j++)
			Turn((Move)genAxis[i]);
		i++;
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void EOPerm::Turn(Move T)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	switch (T)
	{	case U1: *this=*this*U;return;
		case R1: *this=*this*R;return;
		case F1: *this=*this*F;return;
		case D1: *this=*this*D;return;
		case L1: *this=*this*L;return;
		case B1: *this=*this*B;return;

		case U2: *this=*this*U*U;return;
		case R2: *this=*this*R*R;return;
		case F2: *this=*this*F*F;return;
		case D2: *this=*this*D*D;return;
		case L2: *this=*this*L*L;return;
		case B2: *this=*this*B*B;return;

		case U3: *this=*this*U*U*U;return;
		case R3: *this=*this*R*R*R;return;
		case F3: *this=*this*F*F*F;return;
		case D3: *this=*this*D*D*D;return;
		case L3: *this=*this*L*L*L;return;
		case B3: *this=*this*B*B*B;return;
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unsigned short EOPerm::Flip()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	return (unsigned short)OIndex();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unsigned short EOPerm::Slice()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	char* pos = new char[N_EDGE];
	{for (char i=0;i<N_EDGE;i++) pos[i]=0;}
	{for (char i=fr;i<=br;i++) pos[perm.data[i]]=1;}//fr,fl,br,bl positions

	char k=4,i=0;
	int s=0;
	while (k>0)
	{	if (pos[i++]==0)
			s+= Cnk((char)(N_EDGE-i),(char)(k-1));
		else k--;
	}
	delete [] pos;
	return s;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int EOPerm::FlipSlice()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	
	return Flip() + N_EFLIP*Slice();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unsigned short EOPerm::SortedSlice()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	char* pos = new char[N_EDGE];
	{for (char i=0;i<N_EDGE;i++) pos[i]=0;}
	{for (char i=fr;i<=br;i++) pos[perm.data[i]]=1;}//fr,fl,br,bl positions

	char k=4,i=0;
	int s=0;
	while (k>0)
	{	if (pos[i++]==0)
			s+= Cnk((char)(N_EDGE-i),(char)(k-1));
		else k--;
	}
	delete [] pos;

	Perm p(4,perm.data);
	return 24*s + p.Index();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void EOPerm::InitCubeSymTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	for (int k=0;k<N_SYM;k++)	
	{	eCubeSym[k]=CubeSym(k);
		eCubeSymInv[k]=CubeSym(symInv[k]);
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void EOPerm::MakeFlipSliceRepTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{		
	ifstream f(FLIPSLICE,ios::binary|ios::nocreate);
	ifstream fsym(FLIPSLICE_SYM,ios::binary|ios::nocreate);
	if (f && fsym)	return;

	cout<<"Generating "<<N_FLIPSLICE<<" flipSlice classes"<<endl;
	int* bitArray = new int[N_EFLIP*N_SLICE/32];
	unsigned int* rep = new unsigned int[N_FLIPSLICE];
	short int* sym= new short int[N_FLIPSLICE];
	int classNumber=0;
	for (int i=0;i<N_EFLIP*N_SLICE/32;i++)
		bitArray[i]=0;
	for (int m=0;m<N_FLIPSLICE;m++)
		sym[m]=0;

	int base, offset;
	for (int j=0;j<N_EFLIP*N_SLICE/32;j++)
	{	unsigned int bitMask=1;
		for (int k=0;k<32;k++)
		{	if ((bitArray[j]&bitMask)==0)//new representant found
			{	rep[classNumber++]=32*j+k;
				
				if (classNumber%10000==0)
					cout<<classNumber<<" of "<<N_FLIPSLICE<<endl;
				EOPerm e(32*j+k);
				for (char s=0;s<N_SYM;s +=3)//only symmetries which preserve UD-Axis
				{	int idx=(eCubeSym[s]*e*eCubeSymInv[s]).FlipSlice();
					if (idx==32*j+k)//representant has symmetry s
					{
						short int bmsk=1;
						bmsk <<= s/3;
						sym[classNumber-1] |=bmsk;
					}
					base=idx/32;
					offset=idx%32;
					int bmask=1;
					bmask <<=offset;
					bitArray[base] |=bmask;
				}
			}
			bitMask <<=1;
		}
	}
	ofstream of(FLIPSLICE,ios::binary);
	of.write((char *)rep,N_FLIPSLICE*sizeof(int));
	ofstream ofsym(FLIPSLICE_SYM,ios::binary);
	ofsym.write((char *)sym,N_FLIPSLICE*sizeof(short int));
	delete [] bitArray;
	delete [] rep;
	delete [] sym;
}		

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm EOPerm::operator *(EOPerm& p)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	EOPerm result;
	for (char i=0;i<N_EDGE;i++)
	{	result.perm.data[i]=p.perm.data[perm.data[i]];
		result.ori.data[i]=(char)((ori.data[i]+p.ori.data[perm.data[i]])%N_FLIP);
	}
	return result;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm EOPerm::Inverse()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	return EOPerm((*this).OPerm::Inverse());
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EOPerm EOPerm::CubeSym(int x)	//0<=x<48,
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  EOPerm result;
	int counter = x%3;//M_URF3 part
	x /=3;
	while (counter-- >0) result=result*M_URF3;

	counter = x%2;//M_F2 part
	x /=2;
	while (counter-- >0) result=result*M_F2;

	counter = x%4;//M_U4 part
	x /=4;
	while (counter-->0) result=result*M_U4;

	counter=x;
	while (counter-->0) result=result*M_LR2;
	return result;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm::COPerm():OPerm(N_CORNER,N_TWIST)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm::COPerm(const CPerm& p):OPerm(p,N_TWIST)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm::COPerm(int pIndex, int oIndex):OPerm(N_CORNER,N_TWIST,pIndex,oIndex)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm::COPerm(const COri& co):OPerm(co)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm::COPerm(const CPerm& p8, const COri& co):OPerm(p8,co)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm::COPerm(const COPerm& cop):OPerm(cop)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm::COPerm(const OPerm& op):OPerm(op)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm COPerm::Turn(Move T)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	switch (T)
	{	case U1: *this=*this*U;break;
		case R1: *this=*this*R;break;
		case F1: *this=*this*F;break;
		case D1: *this=*this*D;break;
		case L1: *this=*this*L;break;
		case B1: *this=*this*B;break;

		case U2: *this=*this*U*U;break;
		case R2: *this=*this*R*R;break;
		case F2: *this=*this*F*F;break;
		case D2: *this=*this*D*D;break;
		case L2: *this=*this*L*L;break;
		case B2: *this=*this*B*B;break;

		case U3: *this=*this*U*U*U;break;
		case R3: *this=*this*R*R*R;break;
		case F3: *this=*this*F*F*F;break;
		case D3: *this=*this*D*D*D;break;
		case L3: *this=*this*L*L*L;break;
		case B3: *this=*this*B*B*B;break;
	}
	return *this;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unsigned short COPerm::Twist()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	return OIndex();
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm COPerm::operator *(COPerm& p)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	COPerm result;
	if (p.ori.data[0]!=3)
	for (char i=0;i<N_CORNER;i++)
	{	result.perm.data[i]=p.perm.data[perm.data[i]];
		result.ori.data[i]=(char)((ori.data[i]+p.ori.data[perm.data[i]])%N_TWIST);
	}
	else
	//p.ori.data[0]==3 means reflection at the RL-plane
	//This object r is no element of the "orientated permutation group"
	//because if x:=a*(b*r) and y:=(a*b)*r we usually have x!=y. y gives
	//the result we usually want, so we must avoid a*(b*r)
	for (char j=0;j<N_CORNER;j++)
	{	result.perm.data[j]=p.perm.data[perm.data[j]];
		result.ori.data[j]=(char)((N_TWIST-ori.data[j])%N_TWIST);
	}
	return result;
}

COPerm COPerm::Inverse()
{	return COPerm((*this).OPerm::Inverse());
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
COPerm COPerm::CubeSym(int x)	//0<=x<48,
//Return the COPermutation which represents the Cube Symmetry with index x
//0:ID, 1:M_URF, 2:M_URF^2
//3:M_F2, 4:M_URF3*M_F2, 5:M_URF3^2*M_F2 etc.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  COPerm result;
	int counter = x%3;//M_URF3 part
	x /=3;
	while (counter-- >0) result=result*M_URF3;

	counter = x%2;//M_F2 part
	x /=2;
	while (counter-- >0) result=result*M_F2;

	counter = x%4;//M_U4 part
	x /=4;
	while (counter-- >0) result=result*M_U4;

	counter=x;//M_LR2 part
	while (counter-- >0) result=result*M_LR2;
	return result;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
char COPerm::CubeSymIdx(const COPerm& cop)
//Check, if a COPermutation is a Cube Symmetry and return the index
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  COPerm p;

	for (int i=0;i<N_SYM;i++)
	{	p = CubeSym(i);
		for (int j=0;j<3;j++)
		{	if (cop.perm.data[j]!=p.perm.data[j]) break;
			if (j==2) return (char)i;
		}
	}
	return -1; //no symmetry at all
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void COPerm::InitSymmetryGroupTables()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{	
	ifstream ifmult(SYM_MULT,ios::binary|ios::nocreate);
	ifstream ifinv(SYM_INV,ios::binary|ios::nocreate);
	if (!ifmult ||!ifinv )
	{	COPerm pLeft,pRight;
		cout<<"Generating "<<SYM_MULT<<" and "<<SYM_INV<<endl;
		for (int i=0;i<N_SYM;i++)
		{  pLeft=CubeSym(i);
			for (int j=0;j<N_SYM;j++)
			{	pRight=CubeSym(j);
				symMult[i][j]=CubeSymIdx(pLeft*pRight);
				if (symMult[i][j]==0) symInv[i]=(char)j;
			}
		}
		ofstream of1(SYM_MULT,ios::binary);
		of1.write(symMult[0],N_SYM*N_SYM);
		ofstream of2(SYM_INV,ios::binary);
		of2.write(symInv,N_SYM);
	}
	else
	{	ifmult.read(symMult[0],N_SYM*N_SYM);
		ifinv.read(symInv,N_SYM);
	}
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void COPerm::InitMoveConjugateTable()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{  ifstream f(MOVE_CONJ,ios::binary|ios::nocreate);

	if (!f)
	{ 	cout<<"Generating "<<MOVE_CONJ<<endl;
		for (int s=0;s<N_SYM;s++)
		{	
			COPerm S=CubeSym(s); 
			for (char T=0;T<N_MOVE;T++)
			{	COPerm p;
				COPerm r=S*p.Turn((Move)T)*S.Inverse();
				for (char t=0;t<N_MOVE;t++)
				{	COPerm test;
					test.Turn((Move)t);
					int isEqual=1;
					for (int i=0;i<N_CORNER;i++)
					if (test.perm.data[i]!=r.perm.data[i])
					{	isEqual=0;break;
					}
					if (isEqual==1)
					{	moveConj[s][T]=t;
						break;
					}
				}
			}
		}
	
		ofstream of(MOVE_CONJ,ios::binary);
		of.write(&moveConj[0][0],N_SYM*N_MOVE);
	}
	else
	f.read(&moveConj[0][0],N_SYM*N_MOVE);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
unsigned short COPerm::Pos()
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
{
	return PIndex();
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//the following lines define the COPermutation and EOPermutation objects, which
//represent the face turns and the solid moves of the cube
//
//xcpData defines the corner permutation by operation X, xcoData defines the
//changes of the corner orientations. The Identy ID could be
//described by IDpData[8]={urf,ufl,ulb,ubr,drb,dfr,dlf,dbl}
//With   EXAMPLEcpData[8]={ubr,ufl,ulb,drb,dfr,urf,dlf,dbl};
//we describe the permutation urf->ubr, ufl->ufl, ulb->ulb, ubr->drb ......
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static char ucpData[8]={ufl,ulb,ubr,urf,drb,dfr,dlf,dbl};//U corner permutation
static char ucoData[8]={0,  0,  0,  0,  0,  0,  0,  0};
static CPerm ucp(ucpData);
static COri uco(ucoData);
COPerm COPerm::U(ucp,uco);

static char rcpData[8]={ubr,ufl,ulb,drb,dfr,urf,dlf,dbl};//R
static char rcoData[8]={1,  0,  0,  2,  1,  2,  0,  0};
static CPerm rcp(rcpData);
static COri rco(rcoData);
COPerm COPerm::R(rcp,rco);

static char fcpData[8]={dfr,urf,ulb,ubr,drb,dlf,ufl,dbl};//F
static char fcoData[8]={2,  1,  0,  0,  0,  1,  2,  0};
static CPerm fcp(fcpData);
static COri fco(fcoData);
COPerm COPerm::F(fcp,fco);

static char dcpData[8]={urf,ufl,ulb,ubr,dbl,drb,dfr,dlf};//D
static char dcoData[8]={0,  0,  0,  0,  0,  0,  0,  0};
static CPerm dcp(dcpData);
static COri dco(dcoData);
COPerm COPerm::D(dcp,dco);

static char lcpData[8]={urf,dlf,ufl,ubr,drb,dfr,dbl,ulb};//L
static char lcoData[8]={0,  2,  1,  0,  0,  0,  1,  2};
static CPerm lcp(lcpData);
static COri lco(lcoData);
COPerm COPerm::L(lcp,lco);

static char bcpData[8]={urf,ufl,dbl,ulb,ubr,dfr,dlf,drb};//B
static char bcoData[8]={0,  0,  2,  1,  2,  0,  0,  1};
static CPerm bcp(bcpData);
static COri bco(bcoData);
COPerm COPerm::B(bcp,bco);

static char u4cpData[8]={ufl,ulb,ubr,urf,dfr,dlf,dbl,drb};//M_U4
static char u4coData[8]={0,  0,  0,  0,  0,  0,  0,  0};
static CPerm u4cp(u4cpData);
static COri u4co(u4coData);
COPerm COPerm::M_U4(u4cp,u4co);

static char urf3cpData[8]={urf,ubr,drb,dfr,dlf,ufl,ulb,dbl};//M_URF3
static char urf3coData[8]={1,  2,  1,  2,  1,  2,  1,  2};
static CPerm urf3cp(urf3cpData);
static COri urf3co(urf3coData);
COPerm COPerm::M_URF3(urf3cp,urf3co);

static char f2cpData[8]={dlf,dfr,drb,dbl,ulb,ufl,urf,ubr};//M_F2
static char f2coData[8]={0,  0,  0,  0,  0,  0,  0,  0};
static CPerm f2cp(f2cpData);
static COri f2co(f2coData);
COPerm COPerm::M_F2(f2cp,f2co);

static char lr2cpData[8]={ufl,urf,ubr,ulb,dbl,dlf,dfr,drb};//M_LR2
static char lr2coData[8]={3,  3,  3,  3,  3,  3,  3,  3};//see COPermutation::operator*
static CPerm lr2cp(lr2cpData);//for special handling of reflections
static COri lr2co(lr2coData);
COPerm COPerm::M_LR2(lr2cp,lr2co);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//xepData defines the edge permutation by operation X, xeoData defines the changes
//of the edge orientations. The Identy ID could be
//described by IDepData[12]={fr,fl,bl,br,uf,ub,db,df,ur,dr,dl,ul};
//With    EXAMPLEepData[12]={ur,fl,bl,dr,uf,ub,db,df,br,fr,dl,ul};
//we describe the permutation fr->ur, fl->fl, bl->bl, br->dr.......
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static char uepData[12]={fr,fl,bl,br,ul,ur,db,df,uf,dr,dl,ub};//U, Up move
static char ueoData[12]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
static EPerm uep(uepData);
static EOri ueo(ueoData);
EOPerm EOPerm::U(uep,ueo);

static char repData[12]={ur,fl,bl,dr,uf,ub,db,df,br,fr,dl,ul};//R, Right move
static char reoData[12]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
static EPerm rep(repData);
static EOri reo(reoData);
EOPerm EOPerm::R(rep,reo);

static char fepData[12]={df,uf,bl,br,fr,ub,db,fl,ur,dr,dl,ul};//F, Front move
static char feoData[12]={1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0};
static EPerm fep(fepData);
static EOri feo(feoData);
EOPerm EOPerm::F(fep,feo);

static char depData[12]={fr,fl,bl,br,uf,ub,dl,dr,ur,db,df,ul};//D, Down move
static char deoData[12]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
static EPerm dep(depData);
static EOri deo(deoData);
EOPerm EOPerm::D(dep,deo);

static char lepData[12]={fr,dl,ul,br,uf,ub,db,df,ur,dr,bl,fl};//L, Left move
static char leoData[12]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
static EPerm lep(lepData);
static EOri leo(leoData);
EOPerm EOPerm::L(lep,leo);

static char bepData[12]={fr,fl,db,ub,uf,bl,br,df,ur,dr,dl,ul};//B, Back move
static char beoData[12]={0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0};
static EPerm bep(bepData);
static EOri beo(beoData);
EOPerm EOPerm::B(bep,beo);

static char u4epData[12]={fl,bl,br,fr,ul,ur,dr,dl,uf,df,db,ub};//M_U4, clockwise rotation
static char u4eoData[12]={1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0};//along UD-axis by 90 degrees
static EPerm u4ep(u4epData);
static EOri u4eo(u4eoData);
EOPerm EOPerm::M_U4(u4ep,u4eo);

static char urf3epData[12]={uf,ub,db,df,ur,dr,dl,ul,fr,fl,bl,br};//M_URF3, clockwise rotation
static char urf3eoData[12]={0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1};// along urf-dbl-axis by 120 degrees
static EPerm urf3ep(urf3epData);
static EOri urf3eo(urf3eoData);
EOPerm EOPerm::M_URF3(urf3ep,urf3eo);

static char f2epData[12]={fl,fr,br,bl,df,db,ub,uf,dl,ul,ur,dr};//M_F2, rotation
static char f2eoData[12]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};//along FB-axis by 180 degreess
static EPerm f2ep(f2epData);
static EOri f2eo(f2eoData);
EOPerm EOPerm::M_F2(f2ep,f2eo);

static char lr2epData[12]={fl,fr,br,bl,uf,ub,db,df,ul,dl,dr,ur};//M_LR2, reflection
static char lr2eoData[12]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};//at the uf-ub-df-plane
static EPerm lr2ep(lr2epData);//Different from the corners
static EOri lr2eo(lr2eoData);//we have no problems with the edges
EOPerm EOPerm::M_LR2(lr2ep,lr2eo);//concerning the orientations
